#include <sys/platform.h>

#include "../common/spi.h"
#include "../common/flash.h"
#include "../target.h"

static int adi_spi_bf70_open(void)
{
	return 0;
}

static int adi_spi_bf70_close(void)
{
	return 0;
}

static int adi_spi_bf70_config(const struct flash_info *fi)
{
	uint32_t config;

	/* Set SPI CLOCK */
	*pREG_SPI2_CLK = 0x4;

	/* Set SPI CONTROL */

	config = ENUM_SPI_CTL_MASTER;

	if (fi->cpha)
		config |= ENUM_SPI_CTL_SCKBEG;
	else
		config |= ENUM_SPI_CTL_SCKMID;

	if (fi->cpol)
		config |= ENUM_SPI_CTL_SCKLO;
	else
		config |= ENUM_SPI_CTL_SCKHI;

	config |= ENUM_SPI_CTL_SIZE08;

	if (fi->lsb_first)
		config |= ENUM_SPI_CTL_LSB_FIRST;
	else
		config |= ENUM_SPI_CTL_MSB_FIRST;

	if (fi->start_on_mosi)
		config |= ENUM_SPI_CTL_STMOSI;
	else
		config |= ENUM_SPI_CTL_STMISO;

	*pREG_SPI2_CTL = config;

	/* Set SPI word count reload registers */

	*pREG_SPI2_RWCR = 0;
	*pREG_SPI2_TWCR = 0;

	return 0;
}

static int adi_spi_bf70_enable(void)
{
	uint32_t config;

	config = *pREG_SPI2_CTL;
	config &= ~ BITM_SPI_CTL_EN;
	config |= ENUM_SPI_CTL_EN;
	*pREG_SPI2_CTL = config;

	return 0;
}

static int adi_spi_bf70_disable(void)
{
	uint32_t config;

	config = *pREG_SPI2_CTL;
	config &= ~ BITM_SPI_CTL_EN;
	config |= ENUM_SPI_CTL_DIS;
	*pREG_SPI2_CTL = config;

	return 0;
}

static void adi_spi_bf70_wait_send_done(void)
{
	/* It's possible when we check REG_SPI0_STAT, the data has just been
	   moved from TFIFO to shift register.  At that time TFIFO is empty,
	   and SPIF is still set (from last transfer and not cleared yet since
	   the first bit has not been shifted out from the shift register).
	   So we can't simply use (TFIFO_EMPTY && SPIF) as the indication of
	   transfer finish.  We need to check for more times until

	   -- it holds continuously in checking 100 times, (100 is a somewhat
		  arbitrarily chosen number).

	   -- or, it goes through TRUE, FALSE, TRUE again.

	   This method was implemented as below.  But now it's replaced by a
	   new and better mothod.

	uint32_t stat;
	int times, was_true;

	times = 0;
	was_true = 0;
	do
	{
		stat = *pREG_SPI2_STAT;

		if ((stat & BITM_SPI_STAT_TFS) == ENUM_SPI_STAT_TFIFO_EMPTY
				&& (stat & BITM_SPI_STAT_SPIF) == ENUM_SPI_STAT_SPIF_HI)
		{
			if (was_true)
				break;

			times++;
		}
		else if (times != 0)
			was_true = 1;
	}
	while (times < 100);


	   A much better way to check if send is done is setting TWC and
	   checking TF bit, as implemented below.  */

	uint32_t stat;
	do
	{
		stat = *pREG_SPI2_STAT;
		if ((stat & BITM_SPI_STAT_TF) == ENUM_SPI_TF_HI)
			break;
	}
	while (1);
	/* clear the transfer finish indication */
	*pREG_SPI2_STAT = ENUM_SPI_TF_HI;
}

static int adi_spi_bf70_send(uint8_t *buf, int count)
{
	uint32_t config;
	int i;
	int word_count;

	word_count = count / 4;
	while (word_count > 0)
	{
		int twc;

		adi_spi_bf70_disable();
		config = *pREG_SPI2_CTL;
		config &= ~ BITM_SPI_CTL_SIZE;
		config |= ENUM_SPI_CTL_SIZE32;
		*pREG_SPI2_CTL = config;
		adi_spi_bf70_enable();

		twc = word_count < BITM_SPI_TWC_VALUE ? word_count : BITM_SPI_TWC_VALUE;
		*pREG_SPI2_TWC = twc;
		*pREG_SPI2_TXCTL = ENUM_SPI_TXCTL_TX_EN | ENUM_SPI_TXCTL_TTI_EN | ENUM_SPI_TXCTL_TWC_EN;

		for (i = 0; i <= twc; i++)
		{
			uint32_t data;

			data = (buf[i * 4] << 24 | buf[i * 4 + 1] << 16
					| buf[i * 4 + 2] << 8 | buf[i * 4 + 3]);
			while (*pREG_SPI2_STAT & BITM_SPI_STAT_TFF)
				;
			*pREG_SPI2_TFIFO = data;
		}

		adi_spi_bf70_wait_send_done();

		*pREG_SPI2_TXCTL = 0;
		word_count -= twc;
	}

	word_count = count % 4;
	if (word_count > 0)
	{
		adi_spi_bf70_disable();
		config = *pREG_SPI2_CTL;
		config &= ~ BITM_SPI_CTL_SIZE;
		config |= ENUM_SPI_CTL_SIZE08;
		*pREG_SPI2_CTL = config;
		adi_spi_bf70_enable();

		*pREG_SPI2_TWC = word_count;
		*pREG_SPI2_TXCTL = ENUM_SPI_TXCTL_TX_EN | ENUM_SPI_TXCTL_TTI_EN | ENUM_SPI_TXCTL_TWC_EN;

		for (i = 0; i < word_count; i++)
		{
			while (*pREG_SPI2_STAT & BITM_SPI_STAT_TFF)
				;
			*pREG_SPI2_TFIFO = buf[count - word_count + i];
		}

		adi_spi_bf70_wait_send_done();

		*pREG_SPI2_TXCTL = 0;
	}

	return 0;
}

static int adi_spi_bf70_recv(uint8_t *buf, int count)
{
	uint32_t config;
	int i = 0;

	if (count >= 4)
	{
		adi_spi_bf70_disable();
		config = *pREG_SPI2_CTL;
		config &= ~ BITM_SPI_CTL_SIZE;
		config |= ENUM_SPI_CTL_SIZE32;
		*pREG_SPI2_CTL = config;
		adi_spi_bf70_enable();

		*pREG_SPI2_RXCTL = ENUM_SPI_RXCTL_RX_EN | ENUM_SPI_RXCTL_RTI_EN;

		for (i = 0; i + 4 <= count; i += 4)
		{
			uint32_t data;

			while (*pREG_SPI2_STAT & BITM_SPI_STAT_RFE)
				;
			data = *pREG_SPI2_RFIFO;

			buf[i] = (data >> 24) & 0xff;
			buf[i + 1] = (data >> 16) & 0xff;
			buf[i + 2] = (data >> 8) & 0xff;
			buf[i + 3] = data & 0xff;
		}

		*pREG_SPI2_RXCTL = 0;
	}

	if (count > 0)
	{
		adi_spi_bf70_disable();
		config = *pREG_SPI2_CTL;
		config &= ~ BITM_SPI_CTL_SIZE;
		config |= ENUM_SPI_CTL_SIZE08;
		*pREG_SPI2_CTL = config;
		adi_spi_bf70_enable();

		*pREG_SPI2_RXCTL = ENUM_SPI_RXCTL_RX_EN | ENUM_SPI_RXCTL_RTI_EN;

		for (; i < count; i++)
		{
			while (*pREG_SPI2_STAT & BITM_SPI_STAT_RFE)
				;
			buf[i] = *pREG_SPI2_RFIFO;
		}

		*pREG_SPI2_RXCTL = 0;
	}

	return 0;
}

static int adi_spi_bf70_recv_until (int (*condition)(const uint8_t))
{
	uint8_t data;
	uint32_t config;

	adi_spi_bf70_disable();
	config = *pREG_SPI2_CTL;
	config &= ~ BITM_SPI_CTL_SIZE;
	config |= ENUM_SPI_CTL_SIZE08;
	*pREG_SPI2_CTL = config;
	adi_spi_bf70_enable();

	*pREG_SPI2_RXCTL = ENUM_SPI_RXCTL_RX_EN | ENUM_SPI_RXCTL_RTI_EN;

	do
	{
		while (*pREG_SPI2_STAT & BITM_SPI_STAT_RFE)
			;
		data = *pREG_SPI2_RFIFO;
	}
	while (!condition(data));

	*pREG_SPI2_RXCTL = 0;

	return 0;
}

static int adi_spi_bf70_set_mode(spi_mode_t mode)
{
	uint32_t config;

	/* TODO check if SPI is quiescent */

	config = *pREG_SPI2_CTL;
	config &= ~ BITM_SPI_CTL_MIOM;

	switch (mode)
	{
	case STANDARD:
		config |= ENUM_SPI_CTL_MIO_DIS;
		break;

	case DUAL_OUTPUT:
	case DUAL_IO:
		config |= ENUM_SPI_CTL_MIO_DUAL;
		break;

	case QUAD_OUTPUT:
	case QUAD_IO:
		config |= ENUM_SPI_CTL_MIO_QUAD;
		break;

	default:
		/* bad mode */
		return -1;
	}

	*pREG_SPI2_CTL = config;

	return 0;
}

static int adi_spi_bf70_select_slave(int slave_no)
{
	/* currently we only handle slave_no == 1 */
	switch (slave_no)
	{
	case 0:
		SPI_UNSELECT_SLAVE;
		break;

	case 1:
		SPI_SELECT_SLAVE;
		break;

	default:
		return -1;
	}

	return 0;
}

const struct spi_ctlr adi_spi_bf70x_ctlr =
{
	adi_spi_bf70_open,
	adi_spi_bf70_close,
	adi_spi_bf70_config,
	adi_spi_bf70_enable,
	adi_spi_bf70_disable,
	adi_spi_bf70_send,
	adi_spi_bf70_recv,
	adi_spi_bf70_recv_until,
	adi_spi_bf70_set_mode,
	adi_spi_bf70_select_slave,
};
